@@ -274,7 +274,7 @@ class GeoNames(Geocoder):
             exception. Set this only if you wish to override, on this call
             only, the value set during the geocoder's initialization.
 
-        :rtype: :class:`geopy.timezone.Timezone`
+        :rtype: :class:`geopy.timezone.Timezone`.
         """
         ensure_pytz_is_installed()
 
